=pod

=head1 NAME

RG::Liu::LibBlast - Perl modules by J. Liu

=head1 SYNOPSIS

use RG::Liu::LibBlast;

=head1 FUNCTIONS

See module source for description.

=over

=item blastdb_info

=item getBlastStat

=item getQueryLength

=item getQueryName

=item hspView

=item hsspDist

=item isBlast

=item open_gz

=item rd_blast_tab

=item rdBlastHsp

=item rdSaf

=back

=head1 VARIABLES

=over

=back

=head1 AUTHOR

J. Liu

=head1 SEE ALSO

=over

=item Main website

L<https://rostlab.informatik.tu-muenchen.de/wiki/index.php/Slashdata:cubicmanpages>

=back

=cut

package      RG::Liu::LibBlast;
require      Exporter;
@ISA       = qw(Exporter);
@EXPORT    = qw(isBlast rdBlastHsp rd_blast_tab);
@EXPORT_OK = qw(getQueryName getQueryLength hsspDist 
		getBlastStat blastdb_info);

#use strict;

#=============================================================
# subroutines for parsing blast files
# getQueryLength() returns the length of query, -1 if error
# hspView() display HSP in 'graphic' way
# rdBlastHsp() returns all HSPs as list of hash
# isBlast() check the integrety of blast file
#=============================================================

sub blastdb_info {
    my ( $db ) = @_;
    my ( $job_id,$file_dbinfo,$cmd_dbinfo,$line,$fh );

    my $exe_fastacmd = '/usr/pub/molbio/blast/fastacmd';
    if ( ! -s $exe_fastacmd or ! -x $exe_fastacmd ) {
	print STDERR "*** fastacmd executable $exe_fastacmd not found\n";
	return undef;
    }

    my $info = { };

    $info->{seq} = 0;
    $info->{letter} = 0;

    $job_id = $ENV{HOSTNAME}."_$$";
    $file_dbinfo = $db."_".$job_id.".dbinfo";

    $cmd_dbinfo = $exe_fastacmd. " -I -d $db -o $file_dbinfo 2>/dev/null";
    system $cmd_dbinfo;
    return $info if ( ! -s $file_dbinfo );

    open ($fh,$file_dbinfo) or warn "cannot open $file_dbinfo:$!";
    while ($line=<$fh>) {
        $line =~ s/,//g;
        if ( $line =~ /^\s+(\d+) sequences; (\d+) total letters/ ) {
            $info->{seq} = $1;
            $info->{letter} = $2;
            last;
        }
    }
    close $fh;

    unlink $file_dbinfo;
    return $info;
}    


sub getBlastStat {
    my ( $file ) = @_;
    my ( $file_open,$fh,$line,$line_cont );
    my $stat = { };
    
    $file_open = &open_gz($file); # deal with gzipped file
    if ( ! defined $file_open ) {
	print STDERR "$file not found\n";
	return undef;
    }

    open ( $fh,$file_open) or die "cannot open $file_open:$!";
    while ($line=<$fh>) {
	chomp $line;
	if ( $line =~ /^BLASTP (.+)/ ) {
	    $stat->{blast_version} = $1;
	} elsif ( $line =~ /^Database: (.+)/ ) {
	    $stat->{db} = $1;
	    $line_cont = <$fh>;
	    if ( $line_cont =~ /^\s+([0-9,]+) sequences; ([0-9,]+)/ ) {
		$stat->{db_seq} = $1;
		$stat->{db_letter} = $2;
	    }
	} elsif ( $line =~ /^effective length of database:\s*(\S+)/ ) {
            $len = $1;
            $len =~ s/\D+//g;
	    $stat->{effective_db_len} = $len;
            last;
        }
#	last if ( $line =~ /^Searching/ );
    }
    close $fh;
    $stat->{effective_db_len} = 0 if ( ! defined $stat->{effective_db_len} );
    return $stat;
}


sub getQueryName {
    my ( $fileBlast,$opt_tab ) = @_;
    my ( $file_open,$fh_blast,$line,$line_cont);
    my $query = "";
    my $query_anno = "";

    $file_open = &open_gz($fileBlast);
    if ( ! defined $file_open ) {
	return undef;
    }

    
    open ($fh_blast, $file_open) or die "cannot open $fileBlast:$!";
    if ( $opt_tab ) {
	while ($line=<$fh_blast>) {
	    chomp $line;
	    if ( $line =~ /^\# Query:\s*(\S+)\s+/ ) {
		$query = $1;
	    } elsif ( $line =~ /^\# Query annotation: (.+)/ ) {
		$query_anno = $1;
	    }
	    last if ( $line !~ /^\#/ );
	}
    } else {
	while ($line=<$fh_blast>) {
	    if ( $line =~ /^Query=\s*(\S+)\s+(.*)/ ) {
		$query = $1;
		$query_anno = $2." ";
		while ($line_cont=<$fh_blast>) {
		    last if ( $line_cont =~ /\([0-9,]+ letters/ );
		    last if ( $line_cont =~ /^Database/ );
		    $line_cont =~ s/\s+$//g;
		    $query_anno .= $line_cont.' ';
		}
		last;
	    }
	}
	$query_anno =~ s/\s+$//g;
    }
    close $fh_blast;
    return ($query,$query_anno);
}



sub getQueryLength {
    my ( $fileBlast,$opt_tab ) = @_;
    
    my ( $file_open,$fh_blast,$line);   
    
    $file_open = &open_gz($fileBlast);
    if ( ! defined $file_open ) {
	return undef;
    }
    my $queryLength = 0;


    open ($fh_blast, $file_open) or die "cannot open $file_open:$!";
    if ( $opt_tab ) {
	while ($line=<$fh_blast>) {
	    last if ( $line !~ /^\#/ );
	    if ( $line =~ /^\# Query length: (\d+)/ ) {
		$queryLength = $1;
		last;
	    } 
	}
    } else {
	while ($line=<$fh_blast>) {
	    next if ( $line !~ /\w+/ );
	    $line =~ s/,//g;
	    if ( $line =~ /^\s+\(([0-9,]+)\s+letters\)/ ) {
		$queryLength = $1;
		#print "xx $queryLength\n";
		last;
	    } 
	    last if ( $line =~ /^Database\:/ );
	}
    }
    close $fh_blast;
    return $queryLength;
}


sub hspView {
    my ( $fileBlast, $eCutoff, $interval ) = @_;
    my ( $ctInterval,$spaces,@overlap,$overlap,$overlapStart,$overlapEnd);
    my ( $queryId,$queryLength,$mark,$ctMarks,$symbol,$i,$hsp,$hsps,$homoId);
    if ( ! -f $fileBlast ) {
	print STDERR "*** ERROR: $fileBlast not found, exit..\n";
	return;
    }
    if ( ! defined $eCutoff ) {
	$eCutoff = 10;
    }

    if ( defined $interval ) {
	if ( $interval < 5 ) {
	    print STDERR "*** ERROR: will not print by the interval of $interval, exit..\n";
	    return;
	}
    } else {
	$interval = 10;
    }

    $symbol = '-';

    $queryId = $fileBlast;
    $queryId =~ s/\..*//g;
    $queryId =~ s/.*\///g;
    
    $queryLength = &getQueryLength($fileBlast);
    if ( $queryLength == -1 ) {
	print STDERR "*** ERROR: Uable to get query length for $fileBlast, exit..\n";
	return;
    }
    print "query=$queryId\tlength=$queryLength\teCutoff=$eCutoff\n\n";

    $ctInterval = int($queryLength/$interval);
    if ( $ctInterval != $queryLength/$interval ) {
	$ctInterval++;
    }
    $ctMarks = int($ctInterval/10);
    printf "%-15s",' ';
    for $i ( 1..$ctMarks ) {
	$mark = $i * $interval * 10;
	printf "%10d",$mark;
    }
    print "\n";

    printf "%-15s",$queryId;
    
    
    for $i ( 1..$ctInterval ) {
	print $symbol;
    }
    print "\n";
    #return;

    $hsps = &rdBlastHsp($fileBlast);
    $spaces = ' 'x$ctInterval;
    #print "xx $spaces xx\n";die;
    foreach $hsp ( @$hsps ) {
	next if ( $hsp->{expect} > $eCutoff ); # skip irrelevant hits
	@overlap = split //,$spaces;
	#print join('',@overlap),"xx\n";die;
	$homoId = $hsp->{nameLine};
	$homoId =~ s/\s+.*//g;
	$homoId = substr($homoId,0,14);
	$overlapStart = int($hsp->{queryStart}/$interval);
	$overlapEnd = int($hsp->{queryEnd}/$interval);
	for $i ( $overlapStart..$overlapEnd ) {
	    $overlap[$i] = $symbol;
	}
	$overlap = join('',@overlap);
	printf "%-15s%s\n",$homoId,$overlap;
    }
    return;
}


sub hsspDist {
    my ( $lali,$gap,$ct_ide ) = @_;
    my ($len,$pide,$exp);

    $len = $lali - $gap;
    $pide = $ct_ide/$len*100;
    if ($len <= 11) {
	return -(100 + $len);
    } elsif ($len > 450) {
	return $pide - 19.5;
    } else {
	$exp = -0.32 * (1 + exp(- $len / 1000));		
	return $pide - (480 * ($len ** $exp));
    }
}

sub rd_blast_tab {
    my ( $fileBlast ) = @_;
    
    my ( $file_open,@tab_fields,$fh_blast,$line,@tmp,$theHsp,$hsps );

    $file_open = &open_gz($fileBlast);
    return undef if ( ! defined $file_open );
	
    @tab_fields = qw(subjId percIdent alignLen mismatch gap
		     queryStart queryEnd subjStart subjEnd expect
		     score subjLength ctIdent ctPositive hssp subjAnno);

    $hsps = [ ];
    open ( $fh_blast, $file_open ) or die "cannot open $file_open:$!";
    while ( $line = <$fh_blast> ) {
	next if ( $line =~ /^\#/ );
	next if ( $line !~ /\w+/ );
	chomp $line;
	$theHsp = { };
	@{$theHsp}{@tab_fields} = (split /\t/,$line)[1..16];
	
	if ( $theHsp->{ctPositive} ) {
	    $theHsp->{percPositive} = int($theHsp->{ctPositive}/$theHsp->{alignLen}*100);
	}
	$theHsp->{subjAnno} = ' ' if ( ! $theHsp->{subjAnno} );
	$theHsp->{nameLine} = $theHsp->{subjId}.' '.$theHsp->{subjAnno};
	push @$hsps, $theHsp;

    }
    close $fh_blast;

    return $hsps;
}

sub rdBlastHsp {
    my ( $fileBlast ) = @_;
    
    my ( $file_open,@hspField, $patternRound,$fh_blast,$ctRound, $isInHit,$f,$theHsp);
    my ($lengthLine,$identityLine,$scoreLine);
    local ( $queryStart,$queryEnd,$subjStart,$subjEnd);
    local ($nameLine,$subjLength,$score,$expect);
    local ($ctIdent,$alignLen,$percIdent,$ctPositive,$percPositive);
    my ($queryLineStart,$queryLineEnd,$subjLineStart,$subjLineEnd);

    my $hsps = [ ];

    $file_open = &open_gz($fileBlast);
    if ( ! defined $file_open ) {
	print STDERR "*** $fileBlast not found\n";
	return undef;
    }

    @hspField = ('nameLine','subjLength','score','expect',
		 'alignLen','ctIdent','percIdent','ctPositive',
		 'percPositive','gap','queryStart',
		 'queryEnd','subjStart','subjEnd');

    # get round of PSI-BLAST
    $patternRound = "'^Results from round'";
    $ctRound = `egrep $patternRound $fileBlast | wc -l`;
    $ctRound =~ s/\D+//g;

    open ( $fh_blast, $file_open ) or die "cannot open $file_open:$!";
    while ( <$fh_blast> ) {
	if ( $ctRound != 0 ) {
	    next if ( $_ !~ /^Results from round $ctRound/ );
	} else {
	    next if ( $_ !~ /^Searching/);
	}
	
	last;
    } 

    $isInHit = 0;
    $queryStart=$queryEnd=$subjStart=$subjEnd=0;
    while ( <$fh_blast> ) {
				# end of alignment section
	if (/^\s*Database\:/ ) {
	    if ( $queryStart != 0 and $subjStart != 0) { # print the last alignment
		foreach $f ( @hspField ) {
		    $theHsp->{$f} = $$f;
		}
		push @$hsps, $theHsp;
	    }
	    last;
	}

	next if ( $_ !~ /\w+/ );	# skip blank lines

	if ( /^\s*\>/ ) {		# header line

					# record info for the last one
	    if ( $queryStart != 0 and $subjStart != 0 ) {
		foreach $f ( @hspField ) {
		    $theHsp->{$f} = $$f;
		}
		push @$hsps, $theHsp;
	    }
	    
	    $theHsp = { };

	    $queryStart=$queryEnd=$subjStart=$subjEnd=0;
		
	    $isInHit = 1;
	    chomp($nameLine = $_); #print "xx nameLine=$nameLine\n";
	    chomp($lengthLine = <$fh_blast>);

	    while ( $lengthLine !~ /\s*Length\s*=\s*(\d+)/ ) {
		$nameLine .= $lengthLine;
		chomp($lengthLine = <$fh_blast>);

	    }
	    $subjLength = $lengthLine; 
	    $subjLength =~ s/\D+//g;
	    $nameLine =~ s/^\s*\>//;
	    
	} else {
	    next if ( ! $isInHit );
	    if ( /^\s*Score/ ) {	# start of a new HSP
		
				# record info for the last one
		if ( $queryStart and $subjStart ) { 
		    foreach $f ( @hspField ) {
			$theHsp->{$f} = $$f;

		    }
		    push @$hsps, $theHsp;
		}
		$theHsp = { };
		
		$scoreLine = $_;
		$scoreLine =~ s/\s+//g;
		if ( $scoreLine =~ /Score=(.*),Expect=([0-9.e-]+)/ ) {
		    $score = $1;
		    $expect = $2;
		    if ( $expect =~ /^e/ ) {
			$expect = '1'.$expect;
		    }
		}
		
		$identityLine = <$fh_blast>;
		$identityLine =~ s/\s+//g;
		$gapEntry = "";
		if ( $identityLine =~ 
		     /Identities=(\d+)\/(\d+)\((\d+)\%\),Positives=(\d+)\/(\d+)\((\d+)\%\)(.*)$/ ) {
		    ($ctIdent,$alignLen,$percIdent,
		     $ctPositive,$percPositive) = ($1,$2,$3,$4,$6);
		    $gapEntry = $7 if ( $7 );
		    if ( $gapEntry =~ /Gaps=(\d+)\/(\d+)/ ) {
			$gap = $1;
		    } else {
			$gap = 0;
		    }
		}
		
		$queryStart=$queryEnd=$subjStart=$subjEnd=0;
	    } else {
		if ( /^Query\s*:\s*(\d+)\D+(\d+)/ ) { 
		    $queryLineStart = $1; 
		    $queryLineEnd = $2;	
		    if ( $queryStart == 0 ) {
			$queryStart = $queryLineStart;
		    }
		    if ( $queryEnd < $queryLineEnd ) {
			$queryEnd = $queryLineEnd;
		    }
		} elsif ( /^Sbjct\s*:\s*(\d+)\D+(\d+)/ ) {
		    $subjLineStart = $1;
		    $subjLineEnd = $2;
		    if ( $subjStart == 0 ) {
			$subjStart = $subjLineStart;
		    }
		    if ( $subjEnd < $subjLineEnd ) {
			$subjEnd = $subjLineEnd;
		    }
		} else {
		    next;
		}
	    }
	}
    }
    close $fh_blast;

    return $hsps;
}




sub isBlast {			# ==================================
				# check the integrety of blast file
				# ==================================
    my ($fileBlast,$istab) = @_;
    my $isBlast = 0;
    my ($file_open,$fh_blast,$line,$has_name,$has_end,$has_matrix,$has_lambda);

    $file_open = &open_gz($fileBlast);
    return 0 if ( ! defined $file_open );
    open ( $fh_blast, $file_open) or die "cannot open $file_open:$!";
    
    if ( $istab ) {
	while ($line=<$fh_blast>) {
	    $has_name = 1 if ($line =~ /^\# Query:/ );
	    #$has_anno = 1 if ($line =~ /^\# Query annotation:/ );
	    $has_end = 1 if ($line =~ /^\/\// );
	}
	$isBlast = ($has_name and $has_end);
    } else {
	while ( $line=<$fh_blast>) {
	    if ( $line =~ /^Matrix: BLOSUM/i) {
		$has_matrix = 1;
	    } elsif ( $line =~ /^Lambda/ ) {
		$has_lambda = 1;

	    }
	}
	if ( $has_matrix ) {
	    $isBlast = 1;
	} elsif ( $has_lambda ) {
	    $isBlast = 1;
	    print STDERR "** $fileBlast has not Matrix line\n";
	} else {
	    $isBlast = 0;
	}
    }
    close $fh_blast;
    return $isBlast;
}


sub rdSaf {
    my ( $safRef ) = @_;
    my ( @saf, $ctQuery,  $nAlign, $nAlignPdb, $safLine, $currentPos);
    my ( $seqLen,  $name, $safSeq,$lenLine,$seqPerLine, $lineBeg);
    my ( $lineEnd, @seq,  $i, $pos, @pdb, $isPdb, $ctPdbReg, $ctPdbRes);
    my ( @pdbEnd, @pdbBeg, $pdbReg,$id);

    #$[ = 1;

    @saf = @$safRef;

    my $errMsg = "";

				# get nAlign info
    $ctQuery = 0;
    $nAlign = $nAlignPdb = 0;
    foreach $safLine ( @saf ) {
	next if ( $safLine !~ /^\w+/ );
	if ( $safLine =~ /^query/ ) {
	    $ctQuery++;
	    last if ( $ctQuery == 2 );
	    next;
	}
	$nAlign++;
	$nAlignPdb++ if ( $safLine =~ /^pdb/ );
    }
    
				# get aaInPdb info

    $currentPos = 0;
    $seqLen = 0;
    foreach $safLine ( @saf ) {
	next if ( $safLine =~ /^\#/ );
	next if ( $safLine !~ /\w+/ );

	chomp $safLine;
	( $name, $safSeq ) = split /\s+/,$safLine,2;
	if ( ! defined $safSeq ) {
	    $errMsg .= "safSeq not defined\n$id\n$safLine\n$safSeq\n";
	}
	$safSeq =~ s/\s+//g;
	$lenLine = length($safSeq);
	
	if ( $name !~ /\|/ or $name eq "query" ) { # query sequence
	    $seqPerLine = $lenLine;
	    $seqLen += $seqPerLine;
	    $lineBeg = $currentPos;
	    $lineEnd = $currentPos + $seqPerLine;
	    $currentPos = $lineEnd;
	    next;
	}
	
	next if ( $safLine !~ /^pdb\|/);
	
	if ( $seqPerLine != $lenLine ) {
	    $errMsg .= "$safLine\nquery and target have diff length\n";
	    return "";
	}
	
	@seq = split //,$safSeq;
	for $i ( 1..$seqPerLine) {
	    $pos = $lineBeg + $i;
	    next if ( defined $pdb[$pos] and $pdb[$pos] == 1 );
	    if ( ! defined $seq[$i] ) {
		$errMsg .= "xx $id\n$safLine\n$safSeq\nseq[$i] not defined\n";
		die;
	    }
	    if ( $seq[$i] ne '.' ) {
		$pdb[$pos] = 1;
	    }
	}
    }
    

    if ( $seqLen == 0 ) {
	$errMsg .= " seqLen is 0, probably no 'query' line\n";
	return "";
    }


    $isPdb = 0;
    $ctPdbReg = 0;
    $ctPdbRes = 0;
    for $i ( 1..$seqLen ) {
	if ( !defined $pdb[$i] ) {
	    $pdb[$i]=0;
	}
	if ( $pdb[$i] == 1 ) {
	    $ctPdbRes++;
	    if ( $isPdb ) {
		next;
	    } else {
		$isPdb = 1;
		$ctPdbReg++;
		$pdbBeg[$ctPdbReg] = $i;
	    }
	} else {
	    if ( $isPdb ) {
		$pdbEnd[$ctPdbReg] = $i - 1;
		$isPdb = 0;
	    } else {
		next;
	    }
	}
    }
    if ( $isPdb ) {
	$pdbEnd[$ctPdbReg]=$seqLen;
    }

    #print "xx $ctPdbReg\n";die;
    $pdbReg = "";
    for $i ( 1..$ctPdbReg ) {
	$pdbReg .= $pdbBeg[$i].'-'.$pdbEnd[$i].',';
    }
    $pdbReg =~ s/,$//;

    return ( $seqLen, $nAlign, $nAlignPdb, $ctPdbRes, $pdbReg );
}


sub open_gz {
    my ( $file ) = @_;
    my ( $file_zip, $file_unzip, $fileopen );

    undef $fileopen;

    $file_unzip = $file;
    if ( $file =~ /\.gz$/ ) {
	$file_unzip =~ s/\.gz$//;
	$file_zip = $file;
    } else {
	$file_zip = $file.'.gz';
    }
    if ( -s $file_unzip ) {
	$fileopen = "< $file_unzip ";
    } elsif ( -s $file_zip ) {
	$fileopen = " gunzip -c $file_zip | ";
    }

    return $fileopen;
}


1;







