=pod

=head1 NAME

RG::Liu::LibHssp - Perl modules by J. Liu

=head1 SYNOPSIS

use RG::Liu::LibHssp;

=head1 FUNCTIONS

See module source for description.

=over

=item calcEntropy

=item getAaFreq

=item getEntropy

=item getHomoStat

=item getNali

=item getProfile

=item getQueryLength

=item hsspDist

=item isHssp

=back

=head1 VARIABLES

=over

=back

=head1 AUTHOR

J. Liu

=head1 SEE ALSO

=over

=item Main website

L<https://rostlab.informatik.tu-muenchen.de/wiki/index.php/Slashdata:cubicmanpages>

=back

=cut

package      RG::Liu::LibHssp;
require      Exporter;
@ISA       = qw(Exporter);
@EXPORT    = qw(getProfile getEntropy getAaFreq getHomoStat getQueryLength);
@EXPORT_OK = qw(hsspDist isHssp getNali );


#=================================================================
# calculate entropy from HSSP file
#=================================================================



sub getEntropy {
    my ( $fileHssp,$freq_bg, $aa_letter ) = @_;
    my ( $dots, $nAlign, $seqLen, $posAA, $posNOCC, $posAlign, @aa, @alignment );
    my ( $line, @nocc, @entropy,$max_ent, @normalized_ent );

    my $sbr = "getEntropy";
    my $fh = "HSSP_$sbr";
    
    $dots = '....:';

    undef @aa;
    undef @alignment;
    undef @nocc;
    undef @entropy;

    #print "HSSP=$fileHssp\n";
    open ( $fh, $fileHssp ) or die "cannot open $fileHssp:$!";
    while ($line=<$fh> ) {
	last if ( $line =~ /^\#\#\s*ALIGNMENTS/ );
	if ( $line=~ /^NALIGN\s+(\d+)/ ) {
	    $nAlign = $1;
	} elsif ( $line=~ /^SEQLENGTH\s+(\d+)/ ) {
	    $seqLen = $1;
	} else {
	    next;
	}
    }

    while ( $line=<$fh> ) {
	last if ( $line=~/^\#\#\s*SEQUENCE PROFILE/ );
	chomp $line;
	if ( $line=~/\s*SeqNo/ ) {		# header, figure out positions
	    $posAA = index ($line, 'AA');
	    $posAlign = index($line, $dots);
	    $posNOCC = index($line, 'NOCC');
	    next;
	}
	
	if ( $line=~/^\s*(\d+)\s*/ ) {	# alignment block
	    $seqNo = $1;
	    $aa[$seqNo] = substr($line, $posAA, 1);
	    if ( length $line < $posAlign ) {
		$alignment[$seqNo] = "";
	    } else {
		$alignment[$seqNo] .= substr($line,$posAlign,70);
	    }

	    $nocc[$seqNo] = substr($line,$posNOCC,4);
	}
    }
    close $fh;

    $max_ent = 0;
    for $i (1..$seqLen) {
	$entropy[$i] = &calcEntropy($nAlign,$nocc[$i],$alignment[$i],$freq_bg,$aa_letter);
	$max_ent = $entropy[$i] if ( $max_ent < $entropy[$i] );
    }
    
    for $i ( 1..$seqLen ) {
	$normalized_ent[$i] = int($entropy[$i]/$max_ent*100);
    }
    #return ( [ @nocc ], [ @entropy ] );
    return ([@entropy], [@normalized_ent]);
}



sub calcEntropy {
    my ( $nAlign, $nocc, $alignment,$freq_bg,$aa_letter ) = @_;
    my (@seq, $s, %ct, $entropy, $freq, $ctGap );
 
    $ctGap = $nAlign + 1 - $nocc;
    @seq = split //,$alignment;
    foreach $s ( @seq ) {
	$ct{$s}++;
    }

    $entropy = 0;
    foreach $aa ( @$aa_letter ) {
	next if ( ! defined $ct{$aa} or $ct{$aa} == 0 );
	$freq = $ct{$aa} / $nAlign;
	$entropy += $freq * log($freq/$freq_bg->{$aa})/log(2);
    }

				# add entropy to each gap position

    if ( $ctGap ) {
	$entropy += $ctGap/$nAlign *  log($ctGap/$nAlign)/log(2);
    }

    $entropy = (-1) * $entropy;
    print STDERR "$nAlign, $nocc,gap=$ctGap, $alignment,ent=$entropy\n";
    #$entropy =~ s/^\-//;
    return $entropy;
   
}

sub getAaFreq {
    my ($fileIn) = @_;
    my ($sum,@tmp,%f);
    my $sbr = "getAaFreq";
    my $fh = "IN_$sbr";

    $sum = 0;
    open ($fh, $fileIn) or die "cannot open frequency input file $fileIn:$!";
    while ( $line=<$fh> ) {
	next if ( $line=~ /^\s*\#/ );
	chomp $line;
	@tmp = split /\t+/,$line;
	$f{$tmp[0]} = $tmp[2];
	$sum += $tmp[2];
    }
    close $fh;
    return {%f};
}


sub getNali {
    my ( $fileHssp ) = @_;
    my ( $line, $nAlign );

    my $sbr = "getNali";
    my $fh = "HSSP_$sbr";

    open ( $fh, $fileHssp ) or die "cannot open $fileHssp:$!";
    while ($line=<$fh> ) {
	last if ( $line =~ /^\#\#\s*ALIGNMENTS/ );
	if ( $line=~ /^NALIGN\s+(\d+)/ ) {
	    $nAlign = $1;
	} 
    }
    close $fh;

    return $nAlign;
}


sub getQueryLength {
    my ( $fileHssp ) = @_;
    my $sbr = "getQueryLength";
    my $fh = "HSSP_$sbr";
    
    my ( $line,$seqLen );

    return undef if (! -s $fileHssp );
    
    undef $seqLen;

    open ( $fh, $fileHssp ) or die "cannot open $fileHssp:$!";
    while ($line=<$fh> ) {
	last if ( $line =~ /^\#\#\s*ALIGNMENTS/ );
	if ( $line=~ /^SEQLENGTH\s+(\d+)/ ) {
	    $seqLen = $1;
	    last;
	} 
    }
    return $seqLen;
}



sub getHomoStat {
    my ( $fileIn ) = @_;
    my $sbr = "getHits";
    my $fhIn = "IN_$sbr";
    my ( $line,$tmpLine,@tmpLine,@fields,%homo,$f,@homos );
    return undef if ( ! -r $fileIn or ! -s $fileIn );

    @fields = qw(id pide psim gap lali lseq2);
    undef @homos;
    open ($fhIn,$fileIn) or die "cannot open $fileIn:$!";
    while ($line=<$fhIn>) {
	next if ( $line !~ /^\#\# PROTEINS :/ );
	last;
    }
    while ($line=<$fhIn>) {
	last if ($line=~ /^\#\#/ );	# insertion section
	last if ( $line=~/^\/\// ); # end of HSSP file

	next if ( $line !~ /^\s*\d+/ );
	$tmpLine = $line;

	$homo{id} = substr($tmpLine,7,12);
	$homo{pide} = substr($tmpLine,28,4);
	$homo{psim} = substr($tmpLine,33,5);
	$homo{ifirst} = substr($tmpLine,39,4);
	$homo{ilast} = substr($tmpLine,44,4);
	$homo{jfirst} = substr($tmpLine,49,4);
	$homo{jlast} = substr($tmpLine,54,4);
	$homo{gap} =  substr($tmpLine,64,4);
	$homo{lali} = substr($tmpLine,59,4);
	$homo{lseq2} = substr($tmpLine,74,5);

	foreach $f ( @fields ) {
	    $homo{$f} =~ s/\s+//g;
	}
	$homo{'pide'} = 100 * $homo{'pide'};
	$homo{'psim'} = 100 * $homo{'psim'};
	
	push @homos,{%homo};
    }
    close $fhIn;
    return [@homos];	
}				


sub getProfile {
    my ( $fileIn ) = @_;
    my $sbr = "getProfile";
    my $fhIn = "IN_$sbr";
    my ( $line,$tmpLine,@tmpLine,@header,$seqNo,$i,%hash );
    return undef if ( ! -r $fileIn or ! -s $fileIn );

    undef %hash;
    undef @header;
    open ($fhIn,$fileIn) or die "cannot open $fileIn:$!";
    while ($line=<$fhIn>) {
	next if ( $line !~ /^\#\# SEQUENCE PROFILE AND ENTROPY/ );
	last;
    }
    while ($line=<$fhIn>) {
	last if ($line=~ /^\#\#/ );	# insertion section
	last if ( $line=~/^\/\// ); # end of HSSP file
	$tmpLine = $line;
	$seqNo = substr($tmpLine,0,6);
	$pdbNo = substr($tmpLine,6,6);
	$tmpLine = substr($tmpLine,12);

	$tmpLine =~ s/^\s+|\s+$//g;
	@tmpLine = split /\s+/,$tmpLine;
	if ( $line=~/^\s+SeqNo/ ) {
	    @header = @tmpLine;
	    next;
	}
	$seqNo =~ s/\s+//g;
	$pdbNo =~ s/\s+//g;
	$hash{$seqNo}{'pdbNo'} = $pdbNo;
	foreach $i ( 0..$#tmpLine ) {
	    if ( ! defined $header[$i] ) {
		die "HSSP file $fileIn: $i th field in PROFILE not found\nline=$_\n";
	    }
	    $hash{$seqNo}{$header[$i]} = $tmpLine[$i];
	}
    }
    close $fhIn;
    return {%hash};	
}				


sub hsspDist {
    my ( $lali,$pide ) = @_;
    my ($exp);

    if ($lali <= 11) {
	return -(100 + $lali);
    } elsif ($lali > 450) {
	return $pide - 19.5;
    } else {
	$exp = -0.32 * (1 + exp(- $lali / 1000));		
	return $pide - (480 * ($lali ** $exp));
    }
}


sub isHssp {			# ==================================
				# check the integrety of HSSP file
				# ==================================
    my ($file_hssp) = @_;
    my $is_hssp = 0;
    my $fh_hssp;
    
    my ($line,$has_hssp_begin,$has_prot_part,$has_align_part,$has_hssp_end);

    return 0 if ( ! -e $file_hssp);
    open ( $fh_hssp, $file_hssp) or die "cannot open $file_hssp:$!";
    while ($line=<$fh_hssp>) {
	if ($line=~ /^HSSP\s+/) {
	    $has_hssp_begin = 1;
	} elsif ( $line =~ /^\#\# PROTEINS :/ ) {
	    $has_prot_part = 1;
	} elsif ( $line =~ /^\#\# ALIGNMENTS / ) {
	    $has_align_part = 1;
	} elsif ( $line =~ /^\/\// ) {
	    $has_hssp_end = 1;
	}
    }
    close $fh_hssp;
    $is_hssp = 
	$has_hssp_begin && $has_prot_part 
	&& $has_align_part && $has_hssp_end;
	
    return $is_hssp;
}

1;
